==> seabirdfilehandler-0.5.0-py3-none-any.whl: seabirdfilehandler/datafiles.py <==
from pathlib import Path
import xmltodict
import pandas as pd
import logging

logger = logging.getLogger(__name__)


class DataFile:
    """Collection of methods for the SeaBird files that feature some kind of
    data table that is represented in a pandas dataframe.

    Parameters
    ----------

    Returns
    -------

    """

    def __init__(
        self,
        path_to_file: Path | str,
        only_header: bool = False,
    ):
        self.path_to_file = Path(path_to_file)
        self.file_name = self.path_to_file.stem
        self.file_dir = self.path_to_file.parent
        self.only_header = only_header
        self.raw_file_data = []  # the text file input
        self.header = []  # the full file header
        self.sbe9_data = []  # device specific information
        self.metadata = {}  # non-SeaBird metadata
        self.metadata_list = []  # unstructured metadata for easier export
        self.data_table_description = []  # the column names and other info
        self.sensor_data = []
        self.sensors = {}  # xml-parsed sensor data
        self.processing_info = []  # everything after the sensor data
        self.data = []  # the data table
        self.file_data = self.raw_file_data  # variable file information
        self.read_file()
        self.metadata = self.structure_metadata(self.metadata_list)
        if len(self.sensor_data) > 0:
            self.sensors = self.sensor_xml_to_flattened_dict(
                "".join(self.sensor_data)
            )

    def __str__(self) -> str:
        return "/n".join(self.file_data)

    def __repr__(self) -> str:
        return str(self.path_to_file.absolute())

    def __eq__(self, other) -> bool:
        return self.file_data == other.file_data

    def read_file(self):
        """Reads and structures all the different information present in the
        file. Lists and Dictionaries are the data structures of choice. Uses
        basic prefix checking to distinguish different header information.

        Parameters
        ----------

        Returns
        -------

        """
        past_sensors = False
        with self.path_to_file.open("r", encoding="latin-1") as file:
            for line in file:
                self.raw_file_data.append(line)
                line_prefix = line[:2]
                if line_prefix == "* ":
                    self.header.append(line)
                    self.sbe9_data.append(line[2:])
                elif line_prefix == "**":
                    self.header.append(line)
                    self.metadata_list.append(line[3:])
                elif line_prefix == "# ":
                    self.header.append(line)
                    if line[2:].strip()[0] == "<":
                        self.sensor_data.append(line[2:])
                        past_sensors = True
                    else:
                        if past_sensors:
                            self.processing_info.append(line[2:])
                        else:
                            self.data_table_description.append(line[2:])
                else:
                    if line.startswith("*END*"):
                        self.header.append(line)
                        if self.only_header:
                            break
                    else:
                        self.data.append(line)

    def sensor_xml_to_flattened_dict(
        self, sensor_data: str
    ) -> list[dict] | dict:
        """Reads the pure xml sensor input and creates a multilevel dictionary,
        dropping the first two dictionaries, as they are single entry only

        Parameters
        ----------

        Returns
        -------

        """
        full_sensor_dict = xmltodict.parse(sensor_data, process_comments=True)
        try:
            sensors = full_sensor_dict["Sensors"]["sensor"]
        except KeyError as error:
            logger.error(f"XML is not formatted as expected: {error}")
            return full_sensor_dict
        else:
            # create a tidied version of the xml-parsed sensor dict
            tidied_sensor_list = []
            for entry in sensors:
                # use comment value as type descriptor
                comment = entry["#comment"]
                split_comment = comment.split(",")
                new_entry = split_comment[1].strip()
                if split_comment[-1] == " 2":
                    new_entry += " 2"
                # remove second-level dict
                calibration_info = list(entry.values())[-1]
                try:
                    new_dict = {
                        "Channel": entry["@Channel"],
                        "SensorName": new_entry,
                        **calibration_info,
                    }
                except TypeError:
                    new_dict = {
                        "Channel": entry["@Channel"],
                        "SensorName": new_entry,
                        "Info": calibration_info,
                    }
                tidied_sensor_list.append(new_dict)
            return tidied_sensor_list

    def structure_metadata(self, metadata_list: list) -> dict:
        """Creates a dictionary to store the metadata that is added by using
        werums dship API.

        Parameters
        ----------
        metadata_list: list :
            a list of the individual lines of metadata found in the file

        Returns
        -------
        a dictionary of the lines of metadata divided into key-value pairs
        """
        out_dict = {}
        for line in metadata_list:
            try:
                (key, val) = line.split("=")
            except ValueError:
                out_dict["text"] = line
            else:
                out_dict[key.strip()] = val.strip()
        return out_dict

    def define_output_path(
        self,
        file_path: Path | str | None = None,
        file_name: str | None = None,
        file_type: str = ".csv",
    ) -> Path:
        """Creates a Path object holding the desired output path.

        Parameters
        ----------
        file_path : Path :
            directory the file sits in (Default value = self.file_dir)
        file_name : str :
            the original file name (Default value = self.file_name)
        file_type : str :
            the output file type (Default = '.csv')
        Returns
        -------
        a Path object consisting of the full path of the new file

        """
        file_path = self.file_dir if file_path is None else file_path
        file_name = self.file_name if file_name is None else file_name
        if file_type[0] != ".":
            file_type = "." + file_type
        return Path(file_path).joinpath(file_name).with_suffix(file_type)

    def to_csv(
        self,
        selected_columns: list | None = None,
        with_header: bool = True,
        output_file_path: Path | str | None = None,
        output_file_name: str | None = None,
    ):
        """Writes a csv from the current dataframe. Takes a list of columns to
        use, a boolean for writing the header and the output file parameters.

        Parameters
        ----------
        selected_columns : list :
            a list of columns to include in the csv
            (Default value = self.df.columns)
        with_header : boolean :
            indicating whether the header shall appear in the output
             (Default value = True)
        output_file_path : Path :
            file directory (Default value = None)
        output_file_name : str :
            original file name (Default value = None)

        Returns
        -------

        """
        selected_columns = (
            self.df.columns if selected_columns is None else selected_columns
        )
        df = self.df[selected_columns].reset_index(drop=True)
        new_file_path = self.define_output_path(
            output_file_path, output_file_name
        )
        if with_header:
            with open(new_file_path, "w") as file:
                for line in self.header:
                    file.write(line)
            df.to_csv(new_file_path, index=False, mode="a")
        else:
            df.to_csv(new_file_path, index=False, mode="w")
        logger.info(f"Wrote file {self.path_to_file} to {new_file_path}.")

    def selecting_columns(
        self,
        list_of_columns: list | str,
        df: pd.DataFrame | None = None,
    ):
        """Alters the dataframe to only hold the given columns.

        Parameters
        ----------
        list_of_columns: list or str : a collection of columns
        df : pandas.Dataframe :
            Dataframe (Default value = None)

        Returns
        -------

        """
        df = self.df if df is None else df
        # ensure that the input is a list, so that isin() can do its job
        if isinstance(list_of_columns, str):
            list_of_columns = [list_of_columns]
        if isinstance(df, pd.DataFrame):
            self.df = df[list_of_columns].reset_index(drop=True)

==> seabirdfilehandler-0.5.2-py3-none-any.whl: seabirdfilehandler/datafiles.py <==
from pathlib import Path
import xmltodict
import pandas as pd
import numpy as np
import logging

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - [%(levelname)s] - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    handlers=[
        logging.FileHandler("filehandler.log"),
        logging.StreamHandler(),
    ],
)

logger = logging.getLogger(__name__)


class DataFile:
    """Collection of methods for the SeaBird files that feature some kind of
    data table that is represented in a pandas dataframe.

    Parameters
    ----------

    Returns
    -------

    """

    def __init__(
        self,
        path_to_file: Path | str,
        only_header: bool = False,
    ):
        self.path_to_file = Path(path_to_file)
        self.file_name = self.path_to_file.stem
        self.file_dir = self.path_to_file.parent
        self.only_header = only_header
        self.raw_file_data = []  # the text file input
        self.header = []  # the full file header
        self.sbe9_data = []  # device specific information
        self.metadata = {}  # non-SeaBird metadata
        self.metadata_list = []  # unstructured metadata for easier export
        self.data_table_description = []  # the column names and other info
        self.sensor_data = []
        self.sensors = {}  # xml-parsed sensor data
        self.processing_info = []  # everything after the sensor data
        self.data = []  # the data table
        self.file_data = self.raw_file_data  # variable file information
        self.read_file()
        self.metadata = self.structure_metadata(self.metadata_list)
        if len(self.sensor_data) > 0:
            self.sensors = self.sensor_xml_to_flattened_dict(
                "".join(self.sensor_data)
            )

    def __str__(self) -> str:
        return "/n".join(self.file_data)

    def __repr__(self) -> str:
        return str(self.path_to_file.absolute())

    def __eq__(self, other) -> bool:
        return self.file_data == other.file_data

    def read_file(self):
        """Reads and structures all the different information present in the
        file. Lists and Dictionaries are the data structures of choice. Uses
        basic prefix checking to distinguish different header information.

        Parameters
        ----------

        Returns
        -------

        """
        past_sensors = False
        with self.path_to_file.open("r", encoding="latin-1") as file:
            for line in file:
                self.raw_file_data.append(line)
                line_prefix = line[:2]
                if line_prefix == "* ":
                    self.header.append(line)
                    self.sbe9_data.append(line[2:])
                elif line_prefix == "**":
                    self.header.append(line)
                    self.metadata_list.append(line[3:])
                elif line_prefix == "# ":
                    self.header.append(line)
                    if line[2:].strip()[0] == "<":
                        self.sensor_data.append(line[2:])
                        past_sensors = True
                    else:
                        if past_sensors:
                            self.processing_info.append(line[2:])
                        else:
                            self.data_table_description.append(line[2:])
                else:
                    if line.startswith("*END*"):
                        self.header.append(line)
                        if self.only_header:
                            break
                    else:
                        self.data.append(line)

    def sensor_xml_to_flattened_dict(
        self, sensor_data: str
    ) -> list[dict] | dict:
        """Reads the pure xml sensor input and creates a multilevel dictionary,
        dropping the first two dictionaries, as they are single entry only

        Parameters
        ----------

        Returns
        -------

        """
        full_sensor_dict = xmltodict.parse(sensor_data, process_comments=True)
        try:
            sensors = full_sensor_dict["Sensors"]["sensor"]
        except KeyError as error:
            logger.error(f"XML is not formatted as expected: {error}")
            return full_sensor_dict
        else:
            # create a tidied version of the xml-parsed sensor dict
            tidied_sensor_list = []
            for entry in sensors:
                # use comment value as type descriptor
                comment = entry["#comment"]
                split_comment = comment.split(",")
                new_entry = split_comment[1].strip()
                if split_comment[-1] == " 2":
                    new_entry += " 2"
                # remove second-level dict
                calibration_info = list(entry.values())[-1]
                try:
                    new_dict = {
                        "Channel": entry["@Channel"],
                        "SensorName": new_entry,
                        **calibration_info,
                    }
                except TypeError:
                    new_dict = {
                        "Channel": entry["@Channel"],
                        "SensorName": new_entry,
                        "Info": calibration_info,
                    }
                tidied_sensor_list.append(new_dict)
            return tidied_sensor_list

    def structure_metadata(self, metadata_list: list) -> dict:
        """Creates a dictionary to store the metadata that is added by using
        werums dship API.

        Parameters
        ----------
        metadata_list: list :
            a list of the individual lines of metadata found in the file

        Returns
        -------
        a dictionary of the lines of metadata divided into key-value pairs
        """
        out_dict = {}
        for line in metadata_list:
            try:
                (key, val) = line.split("=")
            except ValueError:
                out_dict["text"] = line
            else:
                out_dict[key.strip()] = val.strip()
        return out_dict

    def define_output_path(
        self,
        file_path: Path | str | None = None,
        file_name: str | None = None,
        file_type: str = ".csv",
    ) -> Path:
        """Creates a Path object holding the desired output path.

        Parameters
        ----------
        file_path : Path :
            directory the file sits in (Default value = self.file_dir)
        file_name : str :
            the original file name (Default value = self.file_name)
        file_type : str :
            the output file type (Default = '.csv')
        Returns
        -------
        a Path object consisting of the full path of the new file

        """
        file_path = self.file_dir if file_path is None else file_path
        file_name = self.file_name if file_name is None else file_name
        if file_type[0] != ".":
            file_type = "." + file_type
        return Path(file_path).joinpath(file_name).with_suffix(file_type)

    def to_csv(
        self,
        data: pd.DataFrame | np.ndarray,
        with_header: bool = True,
        output_file_path: Path | str | None = None,
        output_file_name: str | None = None,
    ):
        """Writes a csv from the current dataframe. Takes a list of columns to
        use, a boolean for writing the header and the output file parameters.

        Parameters
        ----------
        selected_columns : list :
            a list of columns to include in the csv
            (Default value = self.df.columns)
        with_header : boolean :
            indicating whether the header shall appear in the output
             (Default value = True)
        output_file_path : Path :
            file directory (Default value = None)
        output_file_name : str :
            original file name (Default value = None)

        Returns
        -------

        """
        new_file_path = self.define_output_path(
            output_file_path, output_file_name
        )
        if with_header:
            with open(new_file_path, "w") as file:
                for line in self.header:
                    file.write(line)
        if isinstance(data, pd.DataFrame):
            data.to_csv(new_file_path, index=False, mode="a")
        else:
            np.savetxt(new_file_path, data, delimiter=",")

    def selecting_columns(
        self,
        list_of_columns: list | str,
        df: pd.DataFrame,
    ):
        """Alters the dataframe to only hold the given columns.

        Parameters
        ----------
        list_of_columns: list or str : a collection of columns
        df : pandas.Dataframe :
            Dataframe (Default value = None)

        Returns
        -------

        """
        # ensure that the input is a list, so that isin() can do its job
        if isinstance(list_of_columns, str):
            list_of_columns = [list_of_columns]
        if isinstance(df, pd.DataFrame):
            self.df = df[list_of_columns].reset_index(drop=True)
